@@ -7,32 +7,7 @@ __copyright__ = "Sean Mullan"
 __license__ = "mit"
 
 
-FULL_RANGE = 0
-TANH = 1
-SIGMOID = 2
-
-
-def _denorm(x, norm_type=None):
-    """Denormalize an image to [0, 255] based on specified or inferred normalization type."""
-    if norm_type is None:
-        if x.max() > 1 and x.max() <= 255 and x.min() >= 0:
-            norm_type = 0  # Values are in range [0, 255], no norm
-        elif x.min() < 0 and x.min() >= -1 and x.max() <= 1:
-            norm_type = 1  # Values are in range [-1, 1], tanh norm
-        elif x.min() >= 0 and x.max() <= 1:
-            norm_type = 2  # Values are in range [0, 1], sigmoid norm
-        else:
-            raise ValueError(
-                "Given values are not in one of the valid image normalization ranges."
-            )
-    if norm_type == TANH:
-        x = ((x * 127.5) + 127.5)
-    elif norm_type == SIGMOID:
-        x = (x * 255.0)
-    return x.astype(np.uint8)
-
-
-def print_grid(image, *images, figsize=(8, 8), toFile=None, show=True, **kwargs):
+def print_grid(*images, figsize=(8, 8), toFile=None, show=True, return_grid_shape=False, **kwargs):
     """Print out images as a grid.
 
     Parameters
@@ -61,80 +36,62 @@ def print_grid(image, *images, figsize=(8, 8), toFile=None, show=True, **kwargs)
         * 2, 3, 4, or 5 dimensional numpy arrays (leading rows will be used as row/column)
     Images can be in shape [x, y] or [x, y, c], but only 1 or 3 channels will work (assumes row/col structure otherwise)
     """
-    if len(images) > 0:
-        images = [image, *images]
-    else:
-        images = image
     defaults = ['hspace', 'wspace', 'left', 'bottom', 'right', 'top']
     for item in defaults:
         if item not in kwargs:
             kwargs[item] = None
-    unwrapped = []
-    if isinstance(images, list):
-        print('list of stuff')
-        if isinstance(images[0], list):
-            # List of lists of images
+
+    if len(images) == 1:
+        images = images[0]
+    if hasattr(images, '__array__'):
+        images = np.array(images)
+    if isinstance(images, (list, tuple)):
+        if isinstance(images[0], (list, tuple)):
+            # list of lists of images
             rows = len(images)
-            cols = max([len(item) for item in images if item is not None])
-            unwrapped = images
-        elif isinstance(images[0], dict):
-            rows = 1
-            cols = len(images)
-            unwrapped = [images]
-        elif isinstance(images[0], np.ndarray) and images[0].ndim in [3, 4] and images[0].shape[-1] not in [1, 3]:
-            # Batches of images - assumes that if image.shape[-1] > 3 that it isn't color channels
-            rows = len(images)
-            cols = max([item.shape[0] for item in images if item is not None])
-            for item in images:
-                unwrapped.append([item[i] if item is not None else None for i in range(item.shape[0])])
-        elif isinstance(images[0], np.ndarray):
-            # Single row of images
-            rows = 1
-            cols = len(images)
-            unwrapped = [images]
+            cols = max([len(item) for item in images])
+            to_show = images
         else:
-            raise ValueError("Error printing image grid")
+            # list of images
+            rows = 1
+            cols = len(images)
+            to_show = [images]
     elif isinstance(images, np.ndarray):
-        if (images.ndim == 3 or images.ndim == 4) and (images.shape[-1] != 3 and images.shape[-1] != 1):
-            if images.ndim == 3:
-                # [batch, x,  y]
-                rows = 1
-                cols = images.shape[0]
-                unwrapped = [images]
-            else:
-                # [row, col, x, y]
-                rows = images.shape[0]
-                cols = images.shape[1]
-                unwrapped = images
-        elif (images.ndim == 3 and images.shape[-1] not in [1, 3]) or images.ndim == 2:
-            # [x, y, c] or [x, y]
+        # input as array
+        while images.shape[0] == 1:
+            images = np.squeeze(images, axis=0)
+        if images.shape[-1] == 1:
+            images = np.squeeze(images, axis=-1)
+        ndim = images.ndim
+        if images.shape[-1] == 3:
+            ndim -= 1
+        if ndim == 2:
+            # single image
             rows = 1
             cols = 1
-            unwrapped = [[images]]
-        elif images.ndim == 4 and (images.shape[-1] == 3 or images.shape[-1] == 1):
-            # [b, x, y, c]
+            to_show = [[images]]
+        elif ndim == 3:
             rows = 1
             cols = images.shape[0]
-            unwrapped = [images]
-        elif images.ndim == 5:
-            # [row, col, x, y, c]
+            to_show = [[images[i] for i in range(images.shape[0])]]
+        elif ndim == 4:
             rows = images.shape[0]
             cols = images.shape[1]
-            unwrapped = images
+            to_show = [[images[i, j] for j in range(images.shape[1])] for i in range(images.shape[0])]
         else:
-            raise ValueError("Invalid array shape {}".format(images.shape))
+            raise ValueError('Invalid array shape: {}'.format(images.shape))
     elif isinstance(images, dict):
-        unwrapped = [[images]]
         rows = 1
         cols = 1
+        to_show = [[images]]
     else:
-        raise TypeError("Print grid requires list/np.ndarray/dict type inputs, not {}".format(type(images)))
+        raise ValueError("Invalid input type: {}".format(type(images)))
 
     fig = plt.figure(figsize=figsize)
     gs = fig.add_gridspec(int(rows), int(cols), hspace=kwargs['hspace'], wspace=kwargs['wspace'])
 
     for row in range(rows):
-        image_row = unwrapped[row]
+        image_row = to_show[row]
         for col in range(len(image_row)):
             image = image_row[col]
             if image is None:
@@ -170,24 +127,67 @@ def print_grid(image, *images, figsize=(8, 8), toFile=None, show=True, **kwargs)
     if 'suptitle' in kwargs:
         plt.suptitle(kwargs['suptitle'])
     if toFile is not None:
-        plt.toFile(toFile, dpi=fig.dpi)
-    if show:
+        plt.savefig(toFile, dpi=fig.dpi)
+    if show:  # pragma: no cover
+        # Check manually
         plt.show()
+    if return_grid_shape:
+        return rows, cols
 
 
-def print_image(image, figsize=(8, 6.5), **kwargs):
-    """Similar to pyplot.imshow, but with axes and margins for a single image."""
-    image = image.flatten()
-    if 'cmap' not in kwargs and image.ndim == 2:
-        kwargs['cmap'] = 'bone'
+def print_image(image, figsize=(8, 6.5), toFile=None, show=True, allow_interpolation=False, imshow_args={}, **kwargs):
+    """Similar to pyplot.imshow, but with axes and margins for a single image.
+
+    Parameters
+    ----------
+    image : numpy.ndarray
+        Input image to display
+    figsize : (float, float)
+        The size in inches of the image to display (the default is (8, 6.5))
+    toFile : str | None
+        The path to save the displayed figure to, or None to not save (the default is None)
+    show : bool
+        Whether to call pyplot.show to display the image (the default is True)
+    allow_interpolation : bool
+        Whether to allow automatic interpolation (nearest neighbor and automatic aspect ratio) when either height or width is 10x larger than the other (the default is False)
+    imshow_args : dict
+        Extra args to pass directly to the pyplot.imshow call (the default is {})
+    """
+    defaults = {
+        'hspace': 0,
+        'wspace': 0,
+        'left': 0,
+        'bottom': 0,
+        'right': 1,
+        'top': 1,
+        'cmap': 'bone'
+    }
+
+    for item in defaults:
+        if item not in kwargs:
+            kwargs[item] = defaults[item]
+    image = np.squeeze(image)
     fig = plt.figure(figsize=figsize)
-    plt.imshow(image.flatten(), cmap=kwargs['cmap'])
+    if max(image.shape[:2]) / min(image.shape[:2]) > 10 and allow_interpolation:
+        imshow_args['interpolation'] = 'nearest'
+        imshow_args['aspect'] = 'auto'
+    if image.ndim == 2:
+        imshow_args['cmap'] = kwargs['cmap']
+
+    plt.imshow(image, **imshow_args)
     plt.gca().set_axis_off()
     plt.subplots_adjust(top=1, bottom=0, right=1, left=0, hspace=0, wspace=0)
+    plt.subplots_adjust(top=kwargs['top'],
+                        bottom=kwargs['bottom'],
+                        right=kwargs['right'],
+                        left=kwargs['left'],
+                        hspace=kwargs['hspace'],
+                        wspace=kwargs['wspace'])
     plt.margins(0, 0)
     plt.gca().xaxis.set_major_locator(plt.NullLocator())
     plt.gca().yaxis.set_major_locator(plt.NullLocator())
-    if 'toFile' in kwargs:
-        plt.savefig(kwargs['toFile'], dpi=fig.dpi)
-    if 'hide' not in kwargs:
+    if toFile is not None:
+        plt.savefig(toFile, dpi=fig.dpi)
+    if show:  # pragma: no cover
+        # Check manually
         plt.show()
